"""Module for gluing the WebUI together."""

from pgfinder import matching, pgio, validation
from pgfinder.gui.internal import (
    MASS_LIB_DIR,
    ms_upload_reader,
    theo_masses_upload_reader,
)


def mass_library_index():
    return open(MASS_LIB_DIR / "index.json").read()


def allowed_modifications():
    return validation.allowed_modifications()


def run_analysis():
    from pyio import (
        cleanupWindow,
        consolidationPpm,
        enabledModifications,
        massLibrary,
        msData,
        ppmTolerance,
    )

    theo_masses = theo_masses_upload_reader(upload=massLibrary.to_py())

    def analyze(virt_file):
        ms_data = ms_upload_reader(virt_file)
        matched = matching.data_analysis(
            ms_data, theo_masses, cleanupWindow, enabledModifications, ppmTolerance, consolidationPpm
        )
        return pgio.dataframe_to_csv_metadata(matched)

    return {f["name"]: analyze(f) for f in msData.to_py()}



"""Fixtures for various tests."""

import datetime
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

import pgfinder.pgio as pgio
import pgfinder.validation as validation

BASE_DIR = Path.cwd()
RESOURCES_DIR = BASE_DIR / "tests" / "resources"
DATA_DIR = RESOURCES_DIR / "data"


# test_matching.py fixtures
@pytest.fixture
def raw_data_filename():
    return str(DATA_DIR / "maxquant_test_data.txt")


@pytest.fixture
def raw_data_no_match_filename():
    return str(DATA_DIR / "maxquant_test_data_no_match.txt")


@pytest.fixture
def theo_masses_filename():
    return str(DATA_DIR / "masses" / "e_coli_monomer_masses.csv")


@pytest.fixture
def raw_data(raw_data_filename):
    my_raw_data = pgio.ms_file_reader(raw_data_filename)
    validation.validate_raw_data_df(my_raw_data)
    return my_raw_data


@pytest.fixture
def raw_data_no_match(raw_data_no_match_filename):
    my_raw_data_no_match = pgio.ms_file_reader(raw_data_no_match_filename)
    validation.validate_raw_data_df(my_raw_data_no_match)
    return my_raw_data_no_match


@pytest.fixture
def theo_masses(theo_masses_filename):
    my_theo_masses = pgio.theo_masses_reader(theo_masses_filename)
    validation.validate_theo_masses_df(my_theo_masses)
    return my_theo_masses


@pytest.fixture
def ipywidgets_upload_output(ftrs_file_name):
    return {
        "name": "ftrs_test_data.ftrs",
        "type": "",
        "size": 14274560,
        "content": open(ftrs_file_name, "rb").read(),
        "last_modified": datetime.datetime(2022, 2, 2, 11, 6, 9, 951000, tzinfo=datetime.timezone.utc),
    }


@pytest.fixture
def ipywidgets_upload_output_theo(theo_masses_file_name):
    return {
        "name": "e_coli_monomer_masses.csv",
        "type": "text/csv",
        "size": 4975,
        "content": open(theo_masses_file_name, "rb").read(),
        "last_modified": datetime.datetime(2022, 3, 13, 19, 47, 12, 102000, tzinfo=datetime.timezone.utc),
    }


@pytest.fixture
def ppm():
    return 10


@pytest.fixture
def obs_theoretical_molecular_weights() -> pd.DataFrame:
    """Provides a simple list of theoretical and observed molecular weights for testing calculate_delta_ppm()."""
    return pd.DataFrame({"mwMonoisotopic": [10, 20, 300], "theo_mwMonoisotopic": [9, "15, 21", "298, 300, 301"]})


# test_regression.py fixtures
@pytest.fixture
def theo_masses_file_name():
    return str(DATA_DIR / "masses" / "e_coli_monomer_masses.csv")


@pytest.fixture
def theo_masses_df(theo_masses_file_name):
    return pgio.theo_masses_reader(theo_masses_file_name)


@pytest.fixture
def mod_test():
    """Modifications used in regression testing."""
    return validation.allowed_modifications()


@pytest.fixture
def mq_file_name():
    return str(DATA_DIR / "maxquant_test_data.txt")


@pytest.fixture
def mq_test_df(mq_file_name):
    return pgio.ms_file_reader(mq_file_name)


@pytest.fixture(params=["ftrs_test_data_5.2.ftrs", "ftrs_test_data_3.11.ftrs"])
def ftrs_file_name(request):
    return str(DATA_DIR / request.param)


@pytest.fixture
def ftrs_test_df(ftrs_file_name):
    return pgio.ms_file_reader(ftrs_file_name)


@pytest.fixture
def sample_df() -> pd.DataFrame:
    """Return a dummy data frame for tests."""
    return pd.DataFrame(
        {
            "id": [1, 1, 2, 2, 3, 4, 4],
            "obs": [1, 5645.35435454, 879546.3924093, 789.3, 6541321.2, 10, 10],
            "exp": [2, 3954.49849514, 879546.8974916, 780.4, 6541329.7, 11, 11],
            "inferred": ["A", "B", "C", "D", "E", "F", "G"],
            "intensity": [2, 2, 3, 3, 5, 6, 6],
        }
    ).convert_dtypes()


@pytest.fixture
def df_diff_ppm(sample_df: pd.DataFrame) -> pd.DataFrame:
    """Return a target data frame for tests with diff_pm included."""
    DELTA_DF = pd.DataFrame(
        {
            "Delta (ppm)": [
                -500000.0,
                427577.82345296827,
                -0.5742528357381609,
                11404.407996,
                -1.299430,
                -90909.09090909091,
                -90909.09090909091,
            ]
        }
    )
    DELTA_DF = pd.concat([sample_df, DELTA_DF], axis=1)
    DELTA_DF = DELTA_DF.convert_dtypes()
    return DELTA_DF[["id", "obs", "exp", "Delta (ppm)", "inferred", "intensity"]]


@pytest.fixture
def df_likely_structure() -> pd.DataFrame:
    """Return a data frame with the lowest ppm differences and their intensity derived."""
    return pd.DataFrame(
        {
            "id": [1, 1, 2, 2, 3, 4, 4],
            "obs": [5645.354355, 1.0, 879546.392409, 789.3, 6541321.2, 10.0, 10.0],
            "exp": [3954.498495, 2.0, 879546.897492, 780.4, 6541329.7, 11.0, 11.0],
            "diff_ppm": [427577.823453, -500000.0, -0.574253, 11404.407996, -1.29943, -90909.090909, -90909.090909],
            "inferred": ["B", "A", "C", "D", "E", "F", "G"],
            "intensity": [2, 1, 3, 4, 5, 6, 6],
            "lowest ppm": [427577.823453, np.nan, -0.574253, np.nan, -1.29943, -90909.090909, -90909.090909],
            "Inferred Max Intensity": [2.0, np.nan, 3.0, np.nan, 5.0, 6.0, 6.0],
        }
    )


@pytest.fixture
def consolidated() -> pd.DataFrame:
    """Return an expected consoldiated dataframe to test against."""
    return pd.DataFrame(
        {
            "id": [1, 2, 3, 4],
            "lowest ppm": ["B", "C", "E", "F,   G"],
            "Inferred Max Intensity": [2.0, 3.0, 5.0, 6.0],
        }
    )


@pytest.fixture
def unconsolidated_df() -> pd.DataFrame:
    """Return a combined unconsolidated and consolidated dataframe to test against."""
    return pd.read_csv(RESOURCES_DIR / "unconsolidated.csv")


@pytest.fixture
def unconsolidated_and_consolidated_df() -> pd.DataFrame:
    """Return a combined unconsolidated and consolidated dataframe to test against."""
    return pd.read_csv(RESOURCES_DIR / "consolidated.csv")



"""Functions for validation of data."""

import io
import logging
import pkgutil
from pathlib import Path

import numpy as np
import pandas as pd

from pgfinder import COLUMNS
from pgfinder.logs.logs import LOGGER_NAME

LOGGER = logging.getLogger(LOGGER_NAME)


def allowed_modifications(config_file: str | Path = "config/allowed_modifications.txt") -> list:
    """
    Loads allowable modifications from a csv file as a list.

    Parameters
    ----------
    config_file : str | Path
        Path to a configuration file. Default 'config/allowed_modifications.txt'

    Returns
    -------
    list
        List of permissible modifications.
    """
    data = pkgutil.get_data(__name__, config_file)
    return [ln.strip().decode("utf-8") for ln in io.BytesIO(data).readlines() if not ln.isspace()]


def validate_raw_data_df(raw_data_df: pd.DataFrame, columns: dict = COLUMNS["pgfinder"]) -> None | ValueError:
    """Validate that the raw data is a Pandas Dataframe with specific column names and that it has attributes.

    Parameters
    ----------
    raw_data_df : pd.DataFrame
        Data frame to be validated.

    Returns
    -------
    None | ValueError
        Error specific to the problem encountered with the data frame if any are encountered.

    Raises
    ------
    ValueError
    """
    if not isinstance(raw_data_df, pd.DataFrame):
        raise ValueError("raw_data_df must be a DataFrame.")

    if not raw_data_df.attrs["file"]:
        raise ValueError("raw_data_df must have a file attribute.")

    colnames = list(columns["input"].values()) + list(columns["inferred"].values())

    if not set(colnames).issubset(set(raw_data_df.columns.to_list())):
        raise ValueError("raw_data_df column names are incorrect")


def validate_theo_masses_df(theo_masses_df: pd.DataFrame) -> None | ValueError:
    """Validate that the theoretical masses data is a Pandas Dataframe with specific column names and that it has
    attributes.

    Parameters
    ----------
    raw_data_df : pd.DataFrame
        Data frame to be validated.

    Returns
    -------
    None | ValueError
        Error specific to the problem encountered with the data frame if any are encountered.

    Raises
    ------
    ValueError
    """

    if not isinstance(theo_masses_df, pd.DataFrame):
        raise ValueError("theo_masses_df must be a DataFrame.")

    if not theo_masses_df.attrs["file"]:
        raise ValueError("theo_masses_df must have a file attribute.")

    colnames = ["Inferred structure", "Theo (Da)"]

    if theo_masses_df.columns.to_list() != colnames:
        raise ValueError("theo_masses_df column names are incorrect")

    coltypes = [np.dtype("O"), np.dtype("float64")]

    if theo_masses_df.dtypes.to_list() != coltypes:
        raise ValueError("theo_masses_df column data types are incorrect")


def validate_rt_window(rt_window: float) -> None | ValueError:
    """Validate that rt_window is an float.

    Parameters
    ----------
    rt_window : int
        Value of rt_window to be validated.

    Returns
    -------
    None | ValueError
        If no error nothing is returned, otherwise a ValueError is raised.

    Raises
    ------
    ValueError
    """

    if not isinstance(rt_window, float):
        raise ValueError("rt_window must be a float.")
    # todo: range?


def validate_enabled_mod_list(enabled_mod_list: list) -> None | ValueError:
    """Validate that enabled_mod_list is a list and modifications are allowed.

    Parameters
    ----------
    enabled_mod_list : int
        Value of enabled_mod_list to be validated.

    Returns
    -------
    None | ValueError
        If no error nothing is returned, otherwise a ValueError is raised.

    Raises
    ------
    ValueError
    """

    if not isinstance(enabled_mod_list, list):
        raise ValueError("enabled_mod_list must be a list.")

    allowed_mods = allowed_modifications()

    if not all(item in allowed_mods for item in enabled_mod_list):
        raise ValueError("Requested modification(s) not recognised.")


def validate_user_ppm(user_ppm: int) -> None | ValueError:
    """Validate that user_ppm is an integer.

    Parameters
    ----------
    user_ppm : int
        Value of user_ppm to be validated.

    Returns
    -------
    None | ValueError
        If no error nothing is returned, otherwise a ValueError is raised.

    Raises
    ------
    ValueError
    """

    if not isinstance(user_ppm, int):
        raise ValueError("user_ppm must be an int.")
    # todo: range?



class UserError(ValueError):
    """A ValueError raised by invalid user input.

    UserError messages are intended for direct consumption by the end-user, so
    the message text should make it clear what's gone wrong and which steps the
    user can take to resolve the problem.
    """



#!/usr/bin/env python3
"""Run pgfinder at the command line."""
import argparse as arg
import ast
import importlib.resources as pkg_resources
import logging
import warnings
from pathlib import Path

import yaml

from pgfinder.errors import UserError
from pgfinder.logs.logs import LOGGER_NAME, setup_logger
from pgfinder.matching import data_analysis
from pgfinder.pgio import (
    dataframe_to_csv_metadata,
    default_filename,
    ms_file_reader,
    read_yaml,
    theo_masses_reader,
)
from pgfinder.utils import update_config

LOGGER = setup_logger()
LOGGER = logging.getLogger(LOGGER_NAME)


def create_parser() -> arg.ArgumentParser:
    """Create a parser for reading options."""
    parser = arg.ArgumentParser(
        description="Process sample. Additional arguments over-ride those in the configuration file."
    )
    parser.add_argument(
        "-c", "--config_file", dest="config_file", required=False, help="Path to a YAML configuration file."
    )
    parser.add_argument("--input_file", dest="input_file", required=False, help="Input File")
    parser.add_argument("--ppm_tolerance", dest="ppm_tolerance", type=float, required=False, help="PPM Toleraance.")
    parser.add_argument(
        "--consolidation_ppm",
        dest="consolidation_ppm",
        type=float,
        required=False,
        help="Maximum absolute ppm distance between consolidated structures.",
    )
    parser.add_argument("--masses_file", dest="masses_file", type=str, required=False, help="Theoretical masses file.")
    parser.add_argument("--time_delta", dest="time_delta", type=int, required=False, help="Time delta.")
    parser.add_argument(
        "--mod_list", dest="mod_list", type=ast.literal_eval, required=False, help="Modifications to include."
    )
    parser.add_argument("--output_dir", dest="output_dir", type=str, required=False, help="Output directory.")
    parser.add_argument("--warnings", dest="warnings", type=str, required=False, help="Whether to ignore warnings.")
    parser.add_argument("--quiet", dest="quiet", type=bool, required=False, help="Supress output.")
    parser.add_argument(
        "--float_format", dest="float_format", type=int, required=False, help="Decimal places in output."
    )

    return parser


def process_file(
    input_file: str | Path,
    masses_file: str | Path,
    mod_list: list,
    ppm_tolerance: float = 10,
    consolidation_ppm: float = 1,
    time_delta: int = 0.5,
    output_dir: str | Path = "./",
    float_format: int = 4,
    to_csv: dict = None,
):
    """Process files

    Parameters
    ----------
    input_file : str | Path
        Mass Spectrometry input file to process.
    masses_file : str | Path
        Input file of known masses.
    mod_list : list
        Modifications to include.
    ppm_tolerance : float
        Parts Per Million tolerance for matching.
    time_delta : int
        Time difference.
    output_dir : str | Path
        Output directory where results are written to.
    float_format : int
       Decimal places to use in CSV files.
    to_csv: dict
       Dictionary of options to pass to pd.to_csv(), primarly used to overwrite existing files.
    """
    input_file = Path(input_file)
    masses_file = Path(masses_file)
    output_dir = Path(output_dir)

    df = ms_file_reader(input_file)

    masses = theo_masses_reader(masses_file)
    LOGGER.info(f"PPM Tolerance                      : {ppm_tolerance}")
    LOGGER.info(f"Time Delta                         : {time_delta}")

    results = data_analysis(
        raw_data_df=df,
        theo_masses_df=masses,
        rt_window=time_delta,
        enabled_mod_list=mod_list,
        ppm_tolerance=ppm_tolerance,
        consolidation_ppm=consolidation_ppm,
    )
    LOGGER.info("Processing complete!")
    filename = default_filename()
    dataframe_to_csv_metadata(
        save_filepath=output_dir,
        output_dataframe=results,
        filename=filename,
        float_format=f"%.{float_format}f",
    )
    LOGGER.info(f"Results with metadata saved to      : {output_dir}/{filename}")


def main():
    """Run processing."""

    try:
        # Parse command line options, load config and update with command line options
        parser = create_parser()
        args = parser.parse_args()
        if args.config_file is not None:
            config = read_yaml(args.config_file)
            LOGGER.info(f"Configuration file loaded from     : {args.config_file}")
        else:
            default_config = pkg_resources.open_text(__package__, "default_config.yaml")
            config = yaml.safe_load(default_config.read())
            LOGGER.info("Default configuration file loaded.")
        config = update_config(config, args)

        # Optionally ignore all warnings or just show deprecation warnings
        if config["warnings"] == "ignore":
            warnings.filterwarnings("ignore")
            LOGGER.info("NB : All warnings have been turned off for this run.")
        elif config["warnings"] == "deprecated":

            def fxn():
                warnings.warn("deprecated", DeprecationWarning, stacklevel=2)

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                fxn()
        if config["quiet"]:
            LOGGER.setLevel("ERROR")

        process_file(
            input_file=config["input_file"],
            masses_file=config["masses_file"],
            ppm_tolerance=config["ppm_tolerance"],
            consolidation_ppm=config["consolidation_ppm"],
            time_delta=config["time_delta"],
            mod_list=config["mod_list"],
            output_dir=config["output_dir"],
            float_format=config["float_format"],
        )
    except UserError as e:
        # Avoid dumping a whole stack-trace if it's the user who's done something wrong
        LOGGER.error(e)


if __name__ == "__main__":
    main()



"""Test utils."""

from decimal import Decimal
from pathlib import Path

from pgfinder.utils import convert_path, dict_to_decimal, update_config


def test_convert_path(tmpdir) -> None:
    """Test path conversion."""
    test_dir = str(tmpdir)
    converted_path = convert_path(test_dir)

    assert isinstance(converted_path, Path)
    assert tmpdir == converted_path


def test_update_config() -> None:
    """Test updating configuration."""
    SAMPLE_CONFIG = {"input_file": "there", "masses_file": "something", "output_dir": "here"}
    NEW_VALUES = {"output_dir": "something new"}
    updated_config = update_config(SAMPLE_CONFIG, NEW_VALUES)

    assert isinstance(updated_config, dict)
    assert updated_config["output_dir"] == Path("something new")


def test_dict_to_decimal() -> None:
    """Test conversion of floats to decimal."""
    SAMPLE_DICT = {"a": 164.1234, "b": {"c": 987.6543}, "d": "a string"}

    decimal_dict = dict_to_decimal(SAMPLE_DICT)

    assert isinstance(decimal_dict, dict)
    assert isinstance(decimal_dict["a"], Decimal)
    assert isinstance(decimal_dict["b"]["c"], Decimal)
    assert isinstance(decimal_dict["d"], str)



"""Utilities"""

import logging
from argparse import Namespace
from decimal import Decimal
from pathlib import Path
from typing import Dict, Union

from pgfinder.logs.logs import LOGGER_NAME

LOGGER = logging.getLogger(LOGGER_NAME)

# FIXME : Set precision, ultimately this _shouldn't_ be handled here but instead moved to find_pg.py and
#         defined in the config.yaml that is read in, dirty hard-coded hack for now to get things working so I can
#         implement the delta_ppm feature.
#
#         HOWEVER : You can't currently set this globally because in clean_up() the .quantize() method is used to round
#         things up to "0.00001" which is currently hard coded. Thus the place to do the tidying is likely under
#         clean_up() but it should be flexible and a user specified option rather than hard coded.
# getcontext().prec = 5


def convert_path(path: Union[str, Path]) -> Path:
    """Ensure path is Path object.

    Parameters
    ----------
    path: Union[str, Path]
        Path to be converted.

    Returns
    -------
    Path
        pathlib Path
    """
    return Path().cwd() if path == "./" else Path(path)


def update_config(config: dict, args: Union[dict, Namespace]) -> Dict:
    """Update the configuration with any arguments

    Parameters
    ----------
    config: dict
        Dictionary of configuration (typically read from YAML file specified with '-c/--config <filename>')
    args: Namespace
        Command line arguments
    Returns
    -------
    Dict
        Dictionary updated with command arguments.
    """
    args = vars(args) if isinstance(args, Namespace) else args

    config_keys = config.keys()
    for arg_key, arg_value in args.items():
        if arg_key in config_keys and arg_value is not None:
            original_value = config[arg_key]
            config[arg_key] = arg_value
            LOGGER.info(f"Updated config config[{arg_key}] : {original_value} > {arg_value} ")
    config["input_file"] = convert_path(config["input_file"])
    config["masses_file"] = convert_path(config["masses_file"])
    config["output_dir"] = convert_path(config["output_dir"])
    return config


def dict_to_decimal(dictionary: dict) -> dict:
    """Recursively convert any floats in a dictionary to Decimal.

    Parameters
    ----------
    dictionary: dict
        Dictionary to be converted.

    Returns
    -------
    dict
        Dictionary with floats converted to Decimal
    """
    for key, value in dictionary.items():
        if isinstance(value, dict):
            dict_to_decimal(value)
        else:
            try:
                dictionary[key] = Decimal(value)
            except:  # noqa: E722
                pass
    return dictionary



"""Package initialisation"""

from importlib.metadata import version
from pkgutil import get_data

import yaml

from pgfinder.utils import dict_to_decimal

PARAMETERS_FILE = "config/parameters.yaml"
PARAMETERS = get_data(__package__, PARAMETERS_FILE)
PARAMETERS = yaml.safe_load(PARAMETERS)
PARAMETERS = dict_to_decimal(PARAMETERS)
COLUMNS_FILE = "config/columns.yaml"
COLUMNS = get_data(__package__, COLUMNS_FILE)
COLUMNS = yaml.safe_load(COLUMNS)
MULTIMERS = PARAMETERS["multimer"]
MOD_TYPE = PARAMETERS["mod_type"]
MASS_TO_CLEAN = PARAMETERS["mass_to_clean"]

release = version("pgfinder")
__version__ = ".".join(release.split("."[:2]))



"""Regression test"""

import pandas as pd

import pgfinder.matching as matching
import pgfinder.pgio as pgio
import pgfinder.validation as validation


def test_matching_mq_baseline(
    regtest, mq_test_df: pd.DataFrame, theo_masses_df: pd.DataFrame, mod_test, tmp_path
) -> None:
    """Test that output of the major function in the module is unchanged."""

    validation.validate_raw_data_df(mq_test_df)
    validation.validate_theo_masses_df(theo_masses_df)

    results = matching.data_analysis(mq_test_df, theo_masses_df, 0.5, mod_test, 10, 1)

    output_filepath = pgio.dataframe_to_csv_metadata(
        save_filepath=tmp_path, output_dataframe=results, filename="output_mq.csv"
    )
    output_df = pd.read_csv(output_filepath, index_col=None)
    # Because versions change based on commits due to setuptools_scm we fix the version so tests are consistent
    output_df.iloc[6, 0] = "version : pgfinder"

    print(output_df.to_csv(index=False), file=regtest)


def test_matching_ftrs_baseline(
    regtest, ftrs_test_df: pd.DataFrame, theo_masses_df: pd.DataFrame, mod_test, tmp_path
) -> None:
    """Test that output of the major function in the module is unchanged."""

    validation.validate_raw_data_df(ftrs_test_df)
    validation.validate_theo_masses_df(theo_masses_df)

    results = matching.data_analysis(ftrs_test_df, theo_masses_df, 0.5, mod_test, 10, 1)

    output_filepath = pgio.dataframe_to_csv_metadata(
        save_filepath=tmp_path, output_dataframe=results, filename="output_ftrs.csv"
    )

    output_df = pd.read_csv(output_filepath, index_col=None)
    # Because versions change based on commits due to setuptools_scm we fix the version so tests are consistent
    output_df.iloc[6, 0] = "version : pgfinder"

    print(output_df.to_csv(index=False), file=regtest)



"""Functions used in uploading files from the WebUI."""

import sys
import tempfile
from contextlib import contextmanager
from pathlib import Path

import pandas as pd

from pgfinder import pgio

# Consider switching to using the import importlib.resources as pkg_resources method of loading files that are included
# as part of the package rather than a plain open()
MASS_LIB_DIR = Path(sys.modules["pgfinder"].__file__).parent / "masses"


def ms_upload_reader(upload: dict) -> pd.DataFrame:
    with uploaded_file(upload) as file:
        return pgio.ms_file_reader(file)


def theo_masses_upload_reader(upload: dict) -> pd.DataFrame:
    """Attempt to load a built-in library if no content was uploaded."""

    if upload["content"] is None:
        upload["content"] = open(MASS_LIB_DIR / upload["name"], "rb").read()

    with uploaded_file(upload) as file:
        return pgio.theo_masses_reader(file)


@contextmanager
def uploaded_file(upload: dict):
    with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tempdir:
        file = Path(tempdir) / upload["name"]
        with open(file, "wb") as f:
            f.write(upload["content"])
        yield file



"""PG Finder I/O operations"""

import logging
import sqlite3
from datetime import datetime
from importlib.metadata import version
from pathlib import Path, PurePath

import numpy as np
import pandas as pd
import yaml
from yaml.error import YAMLError

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

from pgfinder import COLUMNS
from pgfinder.errors import UserError
from pgfinder.logs.logs import LOGGER_NAME

LOGGER = logging.getLogger(LOGGER_NAME)


def ms_file_reader(file: str | Path) -> pd.DataFrame:
    """
    Read mass spec data.

    Parameters
    ----------
    file : str | Path
        Path to be loaded.

    Returns
    -------
    pd.DataFrame
        File loaded as Pandas Dataframe.
    """
    # If we get a path, we need to convert to a string for `in` to work
    filename = PurePath(file)

    if filename.suffix == ".ftrs":
        return_df = ftrs_reader(file)
    elif filename.suffix == ".txt":
        return_df = maxquant_file_reader(file)
    else:
        raise UserError(
            (
                "The supplied data file was neither a .ftrs nor a .txt file. Please ensure that "
                "you've selected a valid Byos (.ftrs) or MaxQuant (.txt) file."
            )
        )

    return_df.attrs["file"] = filename.name
    LOGGER.info(f"Mass spectroscopy file loaded from : {filename.name}")
    return return_df


def ftrs_reader(file: str | Path, columns: dict = COLUMNS) -> pd.DataFrame:
    """
    Reads Features file from Byos.

    Parameters
    ----------
    file : str | Path
        Feature file to be read.
    columns : dict
        Dictionary of columns, this defaults to the global COLUMNS which is read from 'config/columns.yaml' and
    simplifies extension to new formats.

    Returns
    -------
    pd.DataFrame
        Pandas DataFrame of features.
    """
    with sqlite3.connect(file) as db:
        sql = "SELECT * FROM Features"
        # Reads sql database into dataframe
        ff = pd.read_sql(sql, db)
        # Adds empty "Inferred structure" and "Theo (Da)" columns
        ff = ff.reindex(columns=[*ff.columns.tolist(), *columns["pgfinder"]["inferred"].values()], fill_value=np.nan)

        # Determine if file is v311 or v52
        is_ftrs_52 = set(columns["ftrs_52"]).issubset(ff.columns)
        is_ftrs_311 = set(columns["ftrs_311"]).issubset(ff.columns)

        if is_ftrs_52:
            ff.rename(
                columns=dict(zip(columns["ftrs_52"], columns["pgfinder"]["input"].values())),
                inplace=True,
            )
        elif is_ftrs_311:
            ff.rename(
                columns=dict(zip(columns["ftrs_311"], columns["pgfinder"]["input"].values())),
                inplace=True,
            )
        else:
            raise UserError(
                "The supplied FTRS file could not be read. Did it come from an unsupported version of Byos?"
            )

        # Reorder columns in dataframe to desired order, dropping unwanted columns
        return _select_and_order_columns(ff)


def _select_and_order_columns(df: pd.DataFrame, columns: dict = COLUMNS) -> pd.DataFrame:
    """
    Select (renamed) columns and order them.

    Parameters
    ----------
    df : pd.DataFrame
        Full dataframe from which a subset of variables is to be returned.
    columns : dict
        Dictionary of columns, this defaults to the global COLUMNS which is read from 'config/columns.yaml' and
    simplifies extension to new formats.


    Returns
    -------
    pd.DataFrame
        Subset of data frame with selected columns in specified order.
    """
    cols_order = list(columns["pgfinder"]["input"].values()) + list(columns["pgfinder"]["inferred"].values())
    # Move Intensity column to the end to match required order
    cols_order.append(cols_order.pop(cols_order.index("Intensity")))
    return df[cols_order].copy()


def theo_masses_reader(file: str | Path) -> pd.DataFrame:
    """
    Reads theoretical masses files (csv) returning a Panda Dataframe

    Parameters
    ----------
    file : str | Path
        Path to file to be loaded.

    Returns
    -------
    pd.DataFrame
        Pandas DataFrame of theoretical masses.
    """
    try:
        cols = ["Structure", "Monoisotopic Mass"]
        theo_masses_df = pd.read_csv(file, usecols=cols)[cols]
        theo_masses_df.columns = ["Inferred structure", "Theo (Da)"]
    except (pd.errors.ParserError, UnicodeDecodeError) as e:
        raise UserError(
            (
                "The supplied mass database doesn't contain valid CSV. Double-check that you've "
                "selected the correct file and that it's plain CSV."
            )
        ) from e
    except pd.errors.EmptyDataError as e:
        raise UserError(
            (
                "The supplied mass database was empty. Double-check that you've "
                "selected the correct file and that it contains CSV data."
            )
        ) from e
    except ValueError as e:
        raise UserError(
            (
                "The supplied mass database didn't have the correct columns. "
                "Have you checked the format of your database against one of the built-in databases?"
            )
        ) from e

    # Check that all structures are followed by "|n" where n is one or more digits
    if not theo_masses_df["Inferred structure"].str.contains(r"\|\d+$").all():
        raise UserError(
            (
                "The supplied mass database contains structures missing the '|n' suffix encoding oligomerisation "
                "state. This should be '|1' for monomers, '|2' for dimers, '|3' for trimers, and so on."
            )
        )
    theo_masses_df.attrs["file"] = PurePath(file).name
    LOGGER.info(f"Theoretical masses loaded from     : {file}")
    return theo_masses_df


def maxquant_file_reader(file: str | Path, columns: dict = COLUMNS):
    """
    Reads maxquant files and outputs data as a dataframe.

    Parameters
    ----------
    filepath : str | Path
        Path to a text file.
    columns : dict
        Dictionary of columns, this defaults to the global COLUMNS which is read from 'config/columns.yaml' and
    simplifies extension to new formats.

    Returns
    -------
    pd.DataFrame
        Pandas Data frame.
    """

    # reads file into dataframe
    try:
        maxquant_df = pd.read_table(file, low_memory=False)
    except pd.errors.EmptyDataError as e:
        raise UserError(
            (
                "No data was found in the supplied .txt file. Have you checked "
                "you're using the allPeptides.txt file from MaxQuant?"
            )
        ) from e
    # adds empty columns for inferred structure and theoretical mass
    maxquant_df = maxquant_df.reindex(
        columns=[*maxquant_df.columns.tolist(), *columns["pgfinder"]["inferred"].values()], fill_value=np.nan
    )

    # insert dataframe index as a column
    maxquant_df.reset_index(level=0, inplace=True)
    # Renames columns to expected column heading required for data_analysis function
    maxquant_df.rename(
        columns=dict(zip(columns["maxquant"], ["ID", "RT (min)", "Obs (Da)"])),
        inplace=True,
    )
    # Reorder columns in dataframe to desired order.
    try:
        return _select_and_order_columns(maxquant_df)
        # maxquant_df = maxquant_df[cols_order]
    except KeyError as e:
        raise UserError(
            (
                "The supplied MaxQuant file could not be read. Have you checked "
                "you're using the allPeptides.txt file from a supported version of MaxQuant?"
            )
        ) from e
    return maxquant_df


def dataframe_to_csv_metadata(
    output_dataframe: pd.DataFrame,
    save_filepath: str | Path = None,
    filename: str | Path = None,
    float_format: str = "%.4f",
) -> str | None:
    """
    Convert dataframe to CSV with metadata.

    If save_filepath is specified return the relative path of the output file, including the filename, otherwise
    return the .csv in the form of a string.

    Parameters
    ----------
    output_dataframe : pd.DataFrame
        Dataframe to output.
    save_filepath : str | Path
        Path to save to.
    filename : str | Path
        Filename to save to.
    float_format : str
        Format for floating point numbers (default 4 decimal places).

    Returns
    -------
    str | None
        Either returns the path to write data to or writes it to CSV.
    """
    release = version("pgfinder")
    _version = ".".join(release.split("."[:2]))

    metadata = [
        f"file : {str(output_dataframe.attrs['file'])}",
        f"masses_file : {str(output_dataframe.attrs['masses_file'])}",
        f"rt_window : {output_dataframe.attrs['rt_window']}",
        f"modifications : {output_dataframe.attrs['modifications']}",
        f"ppm : {output_dataframe.attrs['ppm']}",
        f"consolidation_ppm : {output_dataframe.attrs['consolidation_ppm']}",
        f"version : {_version}",
    ]
    # Add Metadata as first column
    output_dataframe = pd.concat([pd.DataFrame({"Metadata": metadata}), output_dataframe], axis=1)
    # Save the file to disk
    if save_filepath:
        filename = filename if filename is not None else default_filename()
        save_filepath = Path(save_filepath)
        save_filepath.mkdir(parents=True, exist_ok=True)
        output_dataframe.to_csv(save_filepath / filename, index=False, float_format=float_format)
        output = str(save_filepath / filename)
    # Store in memory as a string for returning to Notebook
    else:
        output = output_dataframe.to_csv(index=False, float_format=float_format)

    return output


def default_filename(prefix: str = "results_") -> str:
    """
    Generate a default filename based on the current date/time.

    Parameters
    ----------
    prefix : str
        String to use as a prefix, default is 'results_'.

    Returns
    -------
    str
        Filename with format 'results_YYYY-MM-DD-hh-mm-ss.csv'.
    """
    now = datetime.now()
    date_time = now.strftime("%Y-%m-%d_%H-%M-%S")
    filename = prefix + date_time + ".csv"

    return filename


def read_yaml(filename: str | Path) -> dict:
    """
    Read a YAML file.

    Parameters
    ----------
    filename : str | Path
        YAML file to read.

    Returns
    -------
    dict
        Dictionary of the file.
    """

    with Path(filename).open() as f:
        try:
            return yaml.load(f, Loader=Loader)
        except YAMLError as exception:
            LOGGER.error(exception)
            return {}



"""Matching functions"""

import logging
import re
from decimal import Decimal

import pandas as pd
from pandas.api.types import is_numeric_dtype

from pgfinder import COLUMNS, MASS_TO_CLEAN, MOD_TYPE, MULTIMERS
from pgfinder.errors import UserError
from pgfinder.logs.logs import LOGGER_NAME

LOGGER = logging.getLogger(LOGGER_NAME)

# Only need pgfinder columns so subset those here to simplify subsequent indexing
COLUMNS = COLUMNS["pgfinder"]


def calc_ppm_tolerance(mw: float, ppm_tol: int = 10) -> float:
    """
    Calculates ppm tolerance value

    Parameters
    ----------
    mw : float
        Molecular weight.
    ppm_tol : int
        PPM tolerance

    Returns
    -------
    float
        ?
    """
    return (mw * ppm_tol) / 1000000


def filtered_theo(ftrs_df: pd.DataFrame, theo_df: pd.DataFrame, user_ppm: int) -> pd.DataFrame:
    """
    Generate list of observed structures from theoretical masses dataframe to reduce search space.

    Parameters
    ----------
    ftrs_df : pd.DataFrame
        Features dataframe.
    theo_df : pd.DataFrame
        Theoretical dataframe.
    user_ppm : int
        User specified Parts Per Million.

    Returns
    -------
    pd.DataFrame
        Dataframe filtered on matches with theoretical masses.
    """
    # Match theoretical structures to raw data to generate a list of observed structures
    matched_df = matching(ftrs_df=ftrs_df, matching_df=theo_df, set_ppm=user_ppm)
    # Create dataframe containing only theo_mwMonoisotopic & inferredStructure columns from matched_df
    filtered_df = matched_df[["Inferred structure", "Theo (Da)"]].copy()
    # Drop all rows with NaN values in the Theo (Da) column
    filtered_df.dropna(subset=["Theo (Da)"], inplace=True)

    # Drop duplicate structures and masses
    filtered_df.drop_duplicates(inplace=True)

    if filtered_df.empty:
        raise UserError("No matches were found for this search. Please check your database or increase mass tolerance.")

    return filtered_df


def multimer_builder(theo_df: pd.DataFrame, multimer_type: str, columns: dict = COLUMNS) -> pd.DataFrame:
    """
    Generate multimers (dimers & trimers) from observed monomers.

    Parameters
    ----------
    theo_df : pd.DataFrame
        Dataframe containing theoretical monomerics structures and their corresponding masses.
    multimer_type : str
        Type of multimers to build.
    columns : dict
        Dictionary of pgfinder columns, loaded by default from 'pgfinder/config/columns.yaml'.

    Returns
    -------
    pd.DataFrame
        Dataframe containing theoretical multimers and their corresponding masses.
    """

    theo_mw = []
    theo_struct = []

    # Perhaps need a better way
    inferred_structure = columns["inferred"]["structure"]

    # Builder sub function - calculates multimer mass and name
    def builder(name, mass, mult_num: int):
        for _, row in theo_df.iterrows():
            if (
                len(row[inferred_structure][: len(row[inferred_structure]) - 2]) > 2
            ):  # Prevent dimer creation using just gm (input format is XX|n) X = letters n = number
                mw = row["Theo (Da)"]
                acceptor = row[inferred_structure][: len(row[inferred_structure]) - 2]
                donor = name
                donor_mw = mass
                theo_mw.append(Decimal(mw) + donor_mw + Decimal("-18.0106"))
                # FIXME: In an ideal world, `-` should actually be `~` here, but Excel will throw
                # a hissy-fit about `~` being an escape character, so that's out of scope for now
                joiner = "-" if "Glycosidic" in multimer_type else "="
                theo_struct.append(acceptor + joiner + donor + "|" + str(mult_num))

    # Call builder subfunction with different arguements based on multimer type selected
    # and calculate multimers based on peptide bond through side chain
    multimer = MULTIMERS[multimer_type]
    LOGGER.info(f"Building features for multimer type : {multimer_type}")
    [builder(molecule, Decimal(features["mass"]), features["mult_num"]) for molecule, features in multimer.items()]

    # converts lists to dataframe
    multimer_df = pd.DataFrame(list(zip(theo_mw, theo_struct)), columns=list(columns["inferred"].values()))
    return multimer_df


def modification_generator(filtered_theo_df: pd.DataFrame, mod_type: str) -> pd.DataFrame:
    """
    Generate modified muropeptides (calculates new mass and add modification tag to structure name).

    Parameters
    ----------
    filtered_theo_df : pd.DataFrame
        Pandas DataFrame of theoretical masses that have been filtered.
    mod_type : str
        Modification type ???.

    Returns
    -------
    pd.DataFrame
        Pandas DataFrame of ???.
    """
    mod_mass = Decimal(MOD_TYPE[mod_type]["mass"])
    # NOTE: This regex extracts the modification abbrevation from the end of its full name / type —
    # it simply extracts the bracketed expression at the end of the line
    mod_abbr = re.search(r"\(.*\)$", mod_type).group(0)

    obs_theo_muropeptides_df = filtered_theo_df.copy()
    # Calculate new mass of modified structure
    obs_theo_muropeptides_df["Theo (Da)"] = obs_theo_muropeptides_df["Theo (Da)"].map(lambda x: Decimal(x) + mod_mass)

    # Add modification tags to structure name — there are some special cases that need handling first!
    # FIXME: Kinda pointless to have a file that the user can use to define custom modifications if
    # we're going to hard-code in special cases anyways? I suppose they can still add their own as
    # long as they don't also want any sort of "special" formatting
    base_structure = obs_theo_muropeptides_df["Inferred structure"]
    # FIXME: Absolutely no validation that these structures make sense or are chemically possible —
    # even modifications like "Loss of GlcNAc" don't guarantee that a `g` character is removed from
    # the structure's name. It just chops off the first character with reckless abandon...
    # NOTE: All of these functions assume (with no guarantee) that structures begin with `gm-`
    special_cases = {
        "Extra Disaccharide (+gm)": lambda s: "gm-" + s,
        "Lactyl Peptides (Lac)": lambda s: "Lac" + s[2:],
        "Loss of Disaccharide (-gm)": lambda s: s[3:],
        "Loss of GlcNAc (-g)": lambda s: s[1:],
    }

    # The silly `len(s) - 2` rubbish here is to preserve the `|x` multimer number at the end of
    # each structure name
    def default_case(s):
        return s[: len(s) - 2] + " " + mod_abbr + s[len(s) - 2 : len(s)]

    structure_updater = special_cases.get(mod_type, default_case)

    obs_theo_muropeptides_df["Inferred structure"] = base_structure.map(structure_updater)
    return obs_theo_muropeptides_df


def matching(ftrs_df: pd.DataFrame, matching_df: pd.DataFrame, set_ppm: int) -> pd.DataFrame:
    """
    Match theoretical masses to observed masses within ppm tolerance.

    Parameters
    ----------
    ftrs_df : pd.DataFrame
        Features DataFrame
    matching_df : pd.DataFrame
        Matching DataFrame
    set_ppm : int

    Returns
    -------
    pd.DataFrame
        Dataframe of matches.
    """
    molecular_weights = matching_df[["Inferred structure", "Theo (Da)"]]
    matches_df = pd.DataFrame()
    for s, m in molecular_weights.itertuples(index=False):
        # FIXME: I'm not sure if it's better to convert everything to float or
        # to convert everthing to Decimal instead
        m = float(m)
        tolerance = calc_ppm_tolerance(m, set_ppm)
        mw_matches = ftrs_df[(ftrs_df["Obs (Da)"] >= m - tolerance) & (ftrs_df["Obs (Da)"] <= m + tolerance)].copy()

        # If we have matches add the structure and molecular weight then append
        if len(mw_matches.index) > 0:
            mw_matches["Inferred structure"] = s
            mw_matches["Theo (Da)"] = round(m, 4)
            matches_df = pd.concat([matches_df, mw_matches])

    # Merge with raw data
    unmatched = ftrs_df[~ftrs_df.index.isin(matches_df.index)]
    return pd.concat([matches_df, unmatched])


def clean_up(ftrs_df: pd.DataFrame, mass_to_clean: Decimal, time_delta: float) -> pd.DataFrame:
    """
    Clean up a DataFrame.

    Parameters
    ----------
    ftrs_df : pd.DataFrame
        Features dataframe?
    mass_to_clean : Decimal
        Mass to be cleaned.
    time_delta: float
        Clean up window.

    Returns
    -------
    pd.DataFrame:
        Tidied Dataframe.
    """
    # Get the type of adduct based on the mass_to_clean (which is a float)
    adducts = {"sodiated": Decimal("21.9819"), "potassated": Decimal("37.9559"), "decay": Decimal("203.0793")}
    adducts_keys = list(adducts.keys())
    adducts_values = list(adducts.values())
    adduct = adducts_keys[adducts_values.index(mass_to_clean)]
    # Selector substrings for generating parent and adduct dataframes
    parent = MASS_TO_CLEAN[adduct]["parent"]
    target = MASS_TO_CLEAN[adduct]["target"]

    # Generate parent dataframe - contains parents
    parent_muropeptide_df = ftrs_df.loc[ftrs_df["Inferred structure"].str.contains(parent, na=False)]

    # Generate adduct dataframe - contains adducts
    adducted_muropeptide_df = ftrs_df.loc[ftrs_df["Inferred structure"].str.contains(target, na=False)]

    # Generate copy of rawdata dataframe
    consolidated_decay_df = ftrs_df.copy()

    # Status updates (prints to console)
    if parent_muropeptide_df.empty:
        LOGGER.info(f"No {parent}  muropeptides found")
    if adducted_muropeptide_df.empty:
        LOGGER.info(f"No {target} found")
    elif mass_to_clean == adducts["sodiated"]:
        LOGGER.info(f"Processing {adducted_muropeptide_df.size} Sodium Adducts")
    elif mass_to_clean == adducts["potassated"]:
        LOGGER.info(f"Processing {adducted_muropeptide_df.size} potassium adducts")
    elif mass_to_clean == adducts["decay"]:
        LOGGER.info(f"Processing {adducted_muropeptide_df.size} in source decay products")

    # Consolidate adduct intensity with parent ions intensity
    for _y, row in parent_muropeptide_df.iterrows():
        # Get retention time value from row
        rt = row["RT (min)"]
        # Get theoretical monoisotopic mass value from row as list of values
        intact_mw = row["Theo (Da)"]

        # Work out rt window
        upper_lim_rt = rt + time_delta
        lower_lim_rt = rt - time_delta

        # Get all adducts within rt window
        ins_constrained_df = adducted_muropeptide_df[
            adducted_muropeptide_df["RT (min)"].between(lower_lim_rt, upper_lim_rt, inclusive="both")
        ]
        if not ins_constrained_df.empty:
            # Loop through each of the adducts in the RT window, the adducts
            # themselves all have structures containing the `target` string
            for _z, ins_row in ins_constrained_df.iterrows():
                ins_mw = ins_row["Theo (Da)"]

                # Compare parent masses to adduct masses
                mass_delta = abs(
                    Decimal(intact_mw).quantize(Decimal("0.00001")) - Decimal(ins_mw).quantize(Decimal("0.00001"))
                )

                # Is the mass delta the same mass as the target `mass_to_clean`?
                # If so, it's the same structure but that gets its charge from
                # the `target` ion instead of a proton as normal. In this case,
                # consolidate the intensities of the parent (H+) and adduct
                # (`target`+) ions so that the parent intensity has all of the
                # adduct intensities added to it
                if mass_delta == mass_to_clean:
                    insDecay_intensity = ins_row["Intensity"]
                    ID = row.ID
                    drop_ID = ins_row.ID
                    # Because long format leads to rows with duplicate IDs, the ["ID"]
                    # of a row is sometimes different from its index in the dataframe.
                    # Because this is sometimes but not always the case, we need this
                    # lookup line:
                    idx = consolidated_decay_df.loc[consolidated_decay_df["ID"] == ID].index[0]
                    # Make sure the row we are trying to consolidate hasn't already
                    # been consolidated and deleted!
                    if not consolidated_decay_df.loc[consolidated_decay_df["ID"] == drop_ID].empty:
                        # Transfer adduct intensity to the parent ion
                        consolidated_decay_df.at[idx, "Intensity"] += insDecay_intensity
                        # Because long format means both IDs and structures can be duplicated,
                        # only ID + structure pairs can be considered unique. Find where IDs
                        # or structures differ and retain only those in the dataframe. This is
                        # the same as *filtering out* rows in which *both* the ID and structure
                        # match the target from ins_row
                        diff_ID = consolidated_decay_df["ID"] != ins_row["ID"]
                        diff_Structure = consolidated_decay_df["Inferred structure"] != ins_row["Inferred structure"]
                        consolidated_decay_df = consolidated_decay_df[diff_ID | diff_Structure]

    return consolidated_decay_df


def data_analysis(
    raw_data_df: pd.DataFrame,
    theo_masses_df: pd.DataFrame,
    rt_window: float,
    enabled_mod_list: list,
    ppm_tolerance: float,
    consolidation_ppm: float,
) -> pd.DataFrame:
    """
    Perform analysis.

    Parameters
    ----------
    raw_data_df : pd.DataFrame
        User data as Pandas DataFrame.
    theo_masses_df : pd.DataFrame
        Theoretical masses as Pandas DataFrame.
    rt_window : float
        Set time window for in-source decay and salt adduct cleanup.
    enabled_mod_list : list
        List of modifications to enable.
    ppm_tolerance : float
        The ppm tolerance used when matching the theoretical masses of structures to observed ions.
    consolidation_ppm : float
        The minimum absolute ppm difference between two matches before one is picked as "most likely" over the other.

    Returns
    -------
    pd.DataFrame
        Dataframe of matches.
    """
    sugar = Decimal("203.0793")
    sodium = Decimal("21.9819")
    potassium = Decimal("37.9559")

    LOGGER.info("Filtering theoretical masses by observed masses")
    obs_monomers_df = filtered_theo(ftrs_df=raw_data_df, theo_df=theo_masses_df, user_ppm=ppm_tolerance)
    # Make sure the enabled_mod_list (if empty), is actually represented by an empty list
    enabled_mod_list = enabled_mod_list or []

    # NOTE: "Multimers" is a semi-magic keyword here. Multimers and modifications are treated
    # differently by most of the code and have their own sections in `parameters.yaml`, but
    # despite this, all of the multimer and modification flags are passed to `data_analysis()`
    # in the same `enabled_mod_list` variable... This is the shortest path to getting something
    # working now.
    multimer_mods = [m for m in enabled_mod_list if "Multimers" in m]
    other_mods = [m for m in enabled_mod_list if m not in multimer_mods]

    def build_multimers(mod):
        LOGGER.info("Building multimers from obs muropeptides")
        theo_multimers_df = multimer_builder(obs_monomers_df, mod)
        LOGGER.info("Filtering theoretical multimers by observed")
        return filtered_theo(raw_data_df, theo_multimers_df, ppm_tolerance)

    obs_theo_df = pd.concat([obs_monomers_df, *(build_multimers(type) for type in multimer_mods)])

    def apply_modification(mod):
        LOGGER.info(f"Generating {mod} variants")
        return modification_generator(obs_theo_df, mod)

    LOGGER.info("Building custom search file")
    master_frame = pd.concat(
        [
            obs_theo_df,
            *(apply_modification(mod) for mod in other_mods),
        ]
    )

    master_frame = master_frame.astype({"Theo (Da)": float})
    LOGGER.info("Matching")
    matched_data_df = matching(raw_data_df, master_frame, ppm_tolerance)
    LOGGER.info("Cleaning data")

    matched_data_df = calculate_ppm_delta(df=matched_data_df)

    cleaned_df = clean_up(ftrs_df=matched_data_df, mass_to_clean=sodium, time_delta=rt_window)
    cleaned_df = clean_up(ftrs_df=cleaned_df, mass_to_clean=potassium, time_delta=rt_window)
    cleaned_data_df = clean_up(ftrs_df=cleaned_df, mass_to_clean=sugar, time_delta=rt_window)

    cleaned_data_df.sort_values(by=["Intensity", "RT (min)"], ascending=[False, True], inplace=True, kind="stable")
    cleaned_data_df.reset_index(drop=True, inplace=True)

    # Apply some post-processing to the results
    most_likely_df = pick_most_likely_structures(cleaned_data_df, consolidation_ppm)
    final_df = consolidate_results(most_likely_df)

    # set metadata
    final_df.attrs["file"] = raw_data_df.attrs["file"]
    final_df.attrs["masses_file"] = theo_masses_df.attrs["file"]
    final_df.attrs["rt_window"] = rt_window
    final_df.attrs["modifications"] = enabled_mod_list
    final_df.attrs["ppm"] = ppm_tolerance
    final_df.attrs["consolidation_ppm"] = consolidation_ppm

    return final_df


def calculate_ppm_delta(
    df: pd.DataFrame,
    observed: str = COLUMNS["input"]["obs"],
    theoretical: str = COLUMNS["inferred"]["mass"],
    diff: str = COLUMNS["delta"],
) -> pd.DataFrame:
    """
    Calculate the difference in Parts Per Million between observed and theoretical masses.

    The PPM difference between observed and theoretical mass is calculated as...

    .. math:: (1000000 * (obs - theor)) / theor

    The function ensures the column is placed after the theoretical mass column to facilitate its use.

    Parameters
    ----------
    df : pd.DataFrame
        Pandas DataFrame of results.
    observed : str
        Variable that defines the observed PPM.
    theoretical : str
        Variable that defines the theoretical PPM.
    diff: str
        Variable to be created that holds the difference in PPM.

    Returns
    -------
    pd.DataFrame
        Pandas DataFrame with difference noted in column diff.

    """
    column_order = list(df.columns)
    theoretical_position = column_order.index(theoretical) + 1
    df.insert(theoretical_position, diff, (1000000 * (df[observed] - df[theoretical])) / df[theoretical])
    LOGGER.info("Difference in PPM calculated.")
    return df


def pick_most_likely_structures(
    df: pd.DataFrame,
    consolidation_ppm: float,
    columns: dict = COLUMNS,
) -> pd.DataFrame:
    """
    Add rows that consolidate ambiguous matches, picking matches with the closest ppm.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame of structures to be processed.
    consolidation_ppm : float
        Minimum Parts Per Million tolerance distinguishing matches.
    columns : dict
        Dictionary of columns, this defaults to the global COLUMNS which is read from 'config/columns.yaml' and
    simplifies extension to new formats.

    Returns
    -------
    pd.DataFrame
        Dataframe of matches within the specified tolerance. Candidates that are not matched are included
        in the file for completeness.
    """

    def add_most_likely_structure(group) -> pd.DataFrame:
        """
        Determines the most likely structure.

        Parameters
        ----------
        group : pd.DataFrame
            DataFrame of structures.

        Returns
        -------
        pd.DataFrame
            Dataframe of most likely values.
        """
        # Sort by lowest absolute ppm first, then break ties with structures (short to long)
        group.sort_values(
            by=[columns["delta"], columns["inferred"]["structure"]],
            ascending=[True, False],
            key=lambda k: abs(k) if is_numeric_dtype(k) else k,
            inplace=True,
            kind="stable",
        )
        group.reset_index(drop=True, inplace=True)

        abs_min_ppm = group[columns["delta"]].loc[0]
        abs_min_intensity = group[columns["input"]["intensity"]].loc[0]

        min_ppm_structure_idxs = abs(abs(abs_min_ppm) - abs(group[columns["delta"]])) < consolidation_ppm
        min_ppm_structures = ",   ".join(group[columns["inferred"]["structure"]].loc[min_ppm_structure_idxs])

        group.at[0, f"Inferred structure ({columns['best_match_suffix']})"] = min_ppm_structures
        group.at[0, f"Intensity ({columns['best_match_suffix']})"] = abs_min_intensity

        return group

    matched_rows = df[df[columns["inferred"]["structure"]].notnull()]
    unmatched_rows = df[df[columns["inferred"]["structure"]].isnull()]

    grouped_df = matched_rows.groupby("ID", as_index=False, sort=False)
    most_likely = grouped_df.apply(add_most_likely_structure)

    merged_df = pd.concat([most_likely, unmatched_rows])

    return merged_df.reset_index(drop=True)


def consolidate_results(
    df: pd.DataFrame,
    intensity_column: str = f"Intensity ({COLUMNS['best_match_suffix']})",
    structure_column: str = f"Inferred structure ({COLUMNS['best_match_suffix']})",
    rt_column: str = COLUMNS["input"]["rt"],
    theo_column: str = COLUMNS["inferred"]["mass"],
    ppm_column: str = COLUMNS["delta"],
    abundance_column: str = COLUMNS["consolidation"]["Abundance (%)"],
    oligomer_column: str = "Oligomerisation",
    total_column: str = COLUMNS["consolidation"]["Total Intensity"],
    columns: dict = COLUMNS,
) -> pd.DataFrame:
    """
    Add a final table of muropeptide structures and their relative abundances

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame of structures to be processed.
    intensity_column : str
        Intensity column.
    structure_column : str
        Structure column.
    rt_column : str
        RT column.
    theo_column : str
        Theoretical Mass column.
    ppm_column : str
        Delta ppm column.
    abundance_column : str
        Abundance column.
    oligomer_column : str
        Oligomer column.
    total_column : str
        Total column.

    Returns
    -------
    pd.DataFrame
        The input dataframe with additional columns containing the consolidated results.
    """

    def pick_from_highest_intensity_instance(column: pd.Series):
        return column[df.loc[column.index, intensity_column].idxmax()]

    consolidated_df = (
        df.groupby(structure_column)
        .agg(
            {
                rt_column: pick_from_highest_intensity_instance,
                intensity_column: "sum",
                theo_column: pick_from_highest_intensity_instance,
                ppm_column: pick_from_highest_intensity_instance,
            }
        )
        .reset_index()
    )
    total_intensity = consolidated_df[intensity_column].sum()
    consolidated_df[abundance_column] = consolidated_df[intensity_column] / total_intensity * 100

    consolidated_df[oligomer_column] = consolidated_df[structure_column].apply(lambda s: s[-1])
    consolidated_df.sort_values(
        by=[oligomer_column, abundance_column], ascending=[True, False], inplace=True, kind="stable", ignore_index=True
    )

    consolidated_df.at[0, total_column] = total_intensity
    consolidated_df = consolidated_df[
        [total_column, structure_column, abundance_column, rt_column, theo_column, ppm_column]
    ]

    consolidated_df[rt_column] = consolidated_df[rt_column].round(2)
    consolidated_df[ppm_column] = consolidated_df[ppm_column].round(1)
    # Rename columns using mapping defined in pgfinder/config/columns.yaml under 'consolidation'
    consolidated_df.rename(columns=columns["consolidation"], inplace=True)

    return pd.concat([df, consolidated_df], axis=1)



"""Test pgio functions."""

from pathlib import Path
from unittest import TestCase

import pandas as pd

from pgfinder import pgio
from pgfinder.gui.internal import ms_upload_reader, theo_masses_upload_reader

BASE_DIR = Path.cwd()
RESOURCES = BASE_DIR / "tests" / "resources"


def test_ms_file_reader_ftrs(ftrs_file_name):
    """Test Mass Spectrometer file reader with ftrs 311 and 52 file formats."""
    assert isinstance(pgio.ms_file_reader(ftrs_file_name), pd.DataFrame)


def test_ms_file_reader_maxquant(mq_file_name):
    """Test Mass Spectrometer file reader with maxuant formats."""
    assert isinstance(pgio.ms_file_reader(mq_file_name), pd.DataFrame)


def test_ms_upload_reader(ipywidgets_upload_output):
    assert isinstance(ms_upload_reader(ipywidgets_upload_output), pd.DataFrame)


def test_theo_masses_upload_reader(ipywidgets_upload_output_theo):
    assert isinstance(theo_masses_upload_reader(ipywidgets_upload_output_theo), pd.DataFrame)


CONFIG = {
    "this": "is",
    "a": "test",
    "yaml": "file",
    "numbers": 123,
    "logical": True,
    "nested": {"something": "else"},
    "a_list": [1, 2, 3],
}


def test_read_yaml() -> None:
    """Test reading of YAML file."""
    sample_config = pgio.read_yaml(RESOURCES / "test.yaml")

    TestCase().assertDictEqual(sample_config, CONFIG)



"""Test the matching process"""

from pathlib import Path

import pandas as pd
import pytest

import pgfinder.matching as matching
from pgfinder.errors import UserError
from pgfinder.matching import calculate_ppm_delta, consolidate_results, pick_most_likely_structures

BASE_DIR = Path.cwd()
RESOURCES = BASE_DIR / "tests" / "resources"


def test_filtered_theo(raw_data, theo_masses, ppm):
    # this crude test just shows that the code runs
    # a better test would check that the data frame returned is correct
    matching.filtered_theo(raw_data, theo_masses, ppm)


def test_filtered_theo_no_match(raw_data_no_match, theo_masses, ppm):
    with pytest.raises(
        UserError,
        match="No matches were found for this search. Please check your database or increase mass tolerance.",
    ):
        matching.filtered_theo(raw_data_no_match, theo_masses, ppm)


def test_calculate_ppm_delta(sample_df: pd.DataFrame, df_diff_ppm: pd.DataFrame) -> None:
    """Test addition of PPM column."""
    pd.testing.assert_frame_equal(calculate_ppm_delta(sample_df, observed="obs", theoretical="exp"), df_diff_ppm)


def test_pick_most_likely_structures() -> None:
    """Test picking the most likely structure based on ppm"""
    long_df = pd.read_csv(RESOURCES / "long_results.csv")
    wide_df = pd.read_csv(RESOURCES / "wide_results.csv")

    reshaped_long_df = pick_most_likely_structures(long_df, 1)

    pd.testing.assert_frame_equal(reshaped_long_df, wide_df, check_dtype=False)


def test_consolidation(unconsolidated_df: pd.DataFrame, unconsolidated_and_consolidated_df: pd.DataFrame) -> None:
    """Test the post-processing structure / intensity consolidation step"""
    pd.testing.assert_frame_equal(consolidate_results(unconsolidated_df), unconsolidated_and_consolidated_df)
